import boto3
import logging

from botocore.exceptions import ClientError
from datetime import datetime
from timestream.batch import batch_records, DEFAULT_MAX_BATCH_SIZE, DEFAULT_MIN_COMMON_DIMENSIONS
from timestream.dimensions import TimestreamDimensions
from timestream.load_result import TimestreamLoadResult
from timestream.record import RejectedTimestreamRecord, TimestreamRecord
from timestream.timestream_reader import TimestreamReader
from timestream.util import is_time_key, is_version_key
from typing import List


class TimestreamLoader(TimestreamReader):
    def __init__(
        self,
        database_name: str = None,
        table_name: str = None,
        session: boto3.Session = None,
        current_time_ms: int = None,
        logger: logging.Logger = None
    ):
        assert database_name, 'Must specify a database_name'
        assert table_name, 'Must specify table_name'
        super().__init__(session, current_time_ms, logger)
        self.__database_name = database_name
        self.__table_name = table_name

    @property
    def database_name(self) -> str:
        return self.__database_name

    @property
    def table_name(self) -> str:
        return self.__table_name

    def does_database_exist(
        self,
        database_name: str = None,
        raise_err: bool = False
    ) -> bool:
        '''Overridden to maintain backwards compatibility'''
        return super().does_database_exist(
            database_name=database_name or self.__database_name,
            raise_err=raise_err
        )

    def does_table_exist(
        self,
        database_name: str = None,
        table_name: str = None,
        raise_err: bool = False
    ) -> bool:
        '''Overridden to maintain backwards compatibility'''
        return super().does_table_exist(
            database_name=database_name or self.__database_name,
            table_name=table_name or self.__table_name,
            raise_err=raise_err
        )

    def load(
        self,
        data: list,
        capture_time: datetime,
        source: str = None,
        upsert: bool = False,
        ignore_parse_error: bool = True,
        max_batch_size: int = DEFAULT_MAX_BATCH_SIZE,
        min_common_dimensions: int = DEFAULT_MIN_COMMON_DIMENSIONS

    ) -> dict:
        '''
        Performs a Timestream load. Validates the input and parses the data into Timestream records.
        Returns the result of the load after writing to Timestream.
        '''
        return self.__load(
            data,
            capture_time,
            source,
            upsert,
            ignore_parse_error,
            False,
            max_batch_size,
            min_common_dimensions
        ).to_dict()

    def load_preview(
        self,
        data: list,
        capture_time: datetime,
        source: str = None,
        upsert: bool = False,
        ignore_parse_error: bool = True,
        max_batch_size: int = DEFAULT_MAX_BATCH_SIZE,
        min_common_dimensions: int = DEFAULT_MIN_COMMON_DIMENSIONS
    ) -> dict:
        '''
        Performs a dry run of a Timestream load, including validating the input values and parsing data
        into Timestream records.
        Returns the result of the load before writing to Timestream.
        '''
        return self.__load(
            data,
            capture_time,
            source,
            upsert,
            ignore_parse_error,
            True,
            max_batch_size,
            min_common_dimensions
        ).to_dict()

    def __load(
        self,
        data: list,
        capture_time: datetime,
        source: str = None,
        upsert: bool = False,
        ignore_parse_error: bool = True,
        preview: bool = False,
        max_batch_size: int = DEFAULT_MAX_BATCH_SIZE,
        min_common_dimensions: int = DEFAULT_MIN_COMMON_DIMENSIONS
    ) -> TimestreamLoadResult:
        self._logger.info(f'Loading data with {len(data)} elements')

        # Parse incoming data into TimestreamRecords
        result = self.__parse_data(data, upsert, ignore_parse_error)
        if preview or result.error or len(result.accepted_records) == 0:
            return result

        # Batch the accepted records by common dimensions
        self._logger.info('starting batching...')
        batches = batch_records(
            result.accepted_records,
            max_batch_size=max_batch_size,
            min_common_dimensions=min_common_dimensions
        )
        assert isinstance(batches, list) and len(batches) > 0, 'Batches must be a non-empty list'
        assert all(len(b.records) > 0 for b in batches), "Not all batches have records (shouldn't be possible)"
        self._logger.info('batching complete')
        # Write records in batches to Timestream
        result = TimestreamLoadResult()
        for b in batches:
            dimensions = b.common_dimensions
            if source:
                dimensions.add_dimension('source', source, overwrite=True)
            if b.records and upsert:
                accepted_records, rejected_records = self.__write_upsert_records(dimensions, b.records)
            elif b.records:
                accepted_records, rejected_records = self.__write_insert_records(dimensions, b.records, capture_time)

            if accepted_records:
                result.add_accepted_records(accepted_records)
            if rejected_records:
                result.add_rejected_records(rejected_records)
                result.error = Exception('Some records were rejected by Timestream')
        return result

    def __write_insert_records(
        self,
        common_dimensions: TimestreamDimensions,
        records: List[TimestreamRecord],
        capture_time: datetime
    ) -> tuple:
        self._logger.info(f'Writing insert records with {len(records)} records')
        assert capture_time, 'A capture time needs to be provided'
        capture_time_ms = self._dt_to_ms(capture_time)
        accepted_records, rejected_records = records, []
        try:
            _ = self._timestream_write_client.write_records(
                DatabaseName=self.__database_name,
                TableName=self.__table_name,
                CommonAttributes={
                    'Dimensions': common_dimensions.loadable(),
                    'MeasureValueType': 'DOUBLE',
                    'Time': str(capture_time_ms),
                    'Version': 1
                },
                Records=[
                    r.loadable(omit_dimensions=common_dimensions)
                    for r in records
                ]
            )
        except ClientError as e:
            self._logger.error(f'Failed to write insert records: {e}')
            accepted_records, rejected_records = self.__process_client_error(e, records)
        return accepted_records, rejected_records

    def __write_upsert_records(
        self,
        common_dimensions: TimestreamDimensions,
        records: List[TimestreamRecord]
    ) -> tuple:
        self._logger.info(f'Writing upsert records with {len(records)} records')
        accepted_records, rejected_records = records, []
        try:
            _ = self._timestream_write_client.write_records(
                DatabaseName=self.__database_name,
                TableName=self.__table_name,
                CommonAttributes={
                    'Dimensions': common_dimensions.loadable(),
                    'MeasureValueType': 'DOUBLE',
                    'Version': self._current_time_ms
                },
                Records=[
                    r.loadable(omit_dimensions=common_dimensions)
                    for r in records
                ]
            )
        except ClientError as e:
            self._logger.error(f'Failed to write upsert records: {e}')
            accepted_records, rejected_records = self.__process_client_error(e, records)
        return accepted_records, rejected_records

    def __process_client_error(self, e: ClientError, records: List[TimestreamRecord]) -> tuple:
        '''
        Processes a boto ClientError from a Timestream write operation. Logs error messages
        for each rejected Timestream record and returns 2 lists:
          1. accepted_records: the records that were successfully written to Timestream
          2. rejected_records: the records that Timestream rejected, and the reasons why
        '''
        rejected_records, rejected_indices = [], []
        for rejection_data in e.response.get('RejectedRecords', []):
            rejected_record = RejectedTimestreamRecord(
                record=records[rejection_data['RecordIndex']],
                error=Exception(rejection_data['Reason'])
            )
            self._logger.error(f'Rejected record:\n{rejected_record}')
            rejected_records.append(rejected_record)
            rejected_indices.append(rejection_data['RecordIndex'])

        accepted_records = [
            records[i] for i in range(len(records))
            if i not in rejected_indices
        ]
        return accepted_records, rejected_records

    def __parse_data(self, data: List[dict], upsert: bool, ignore_parse_error: bool) -> TimestreamLoadResult:
        '''
        Parses the given data and returns a TimestreamLoadResult that contains:
          1. Successfully parsed TimestreamRecords
          2. Invalid items that failed to parse into TimestreamRecords (and the errors for why they failed to parse)
        '''
        # TODO: detect duplicates
        result = TimestreamLoadResult()
        if not data:
            return result
        assert type(data) is list, f'Invalid data, expecting a list but got type "{type(data)}"'

        for item in data:
            try:
                records_from_item = self.__parse_item(item, upsert, ignore_parse_error)
            except AssertionError as e:
                result.error = Exception('Some items could not be parsed into Timestream records')
                result.add_invalid_item(item, e)
                continue
            result.add_accepted_records(records_from_item)
        return result

    def __parse_item(self, item: dict, upsert: bool, ignore_parse_error: bool) -> List[TimestreamRecord]:
        '''
        Parses a single dictionary into Timestream records. A single dictionary may break down into
        multiple Timestream records if there are multiple numerical values in the dictionary.

        Raises an AssertionError if the item (dict) cannot be parsed into Timestream records.
        If "ignore_parse_error" is true, this will attempt to ignore all values of the item
        that cannot be converted into either a Timestream record or dimension (rather than
        raising an AssertionError).
        '''
        if not item:
            return []
        assert type(item) is dict, f'Invalid item with type "{type(item)}", expecting a dictionary'

        records, dimensions = [], TimestreamDimensions()

        # Get "time" from the item
        mstimes = [v for k, v in item.items() if is_time_key(k)]
        mstime = None if not mstimes else mstimes[0]
        if upsert:
            assert mstime is not None, f'To perform an upsert, item must have a "time" field'
            assert len(mstimes) < 2, f'Item has multiple time fields'
        else:
            assert mstime is None, f'To perform an insert, item cannot specify a "time" field'

        # Build records and dimensions
        for k, v in item.items():
            if v is None or is_time_key(k):
                continue
            elif is_version_key(k):
                raise AssertionError('Field "version" is a protected field and cannot exist on an item')
            elif type(v) in [float, int, list]:
                # Ints, floats, and lists get converted into records
                records.append(TimestreamRecord(
                    measure_name=k,
                    measure_value=v if type(v) in [float, int] else len(v),
                    mstime=mstime
                ))
            elif type(v) is str:
                # Strings get converted into dimensions
                dimensions.add_dimension(k, v)
            else:
                msg = f'Could not create a record or dimension from type "{type(v)}" at key {k}'
                if not ignore_parse_error:
                    raise AssertionError(msg)
                self._logger.warning(msg)

        # Recreate all records with dimensions attached and return
        return [
            TimestreamRecord(
                measure_name=r.measure_name,
                measure_value=r.measure_value,
                dimensions=dimensions,
                mstime=r.mstime
            )
            for r in records
        ]



import pytest
import pytz

from botocore.exceptions import ClientError
from datetime import date, datetime
from test_utils.mock import MockApi
from timestream.api import TimestreamLoader as ApiTimestreamLoader
from timestream.timestream_reader import DEFAULT_WRITE_CLIENT_CONFIG
from timestream.timestream_loader import TimestreamLoader
from timestream.dimensions import TimestreamDimensions
from timestream.load_result import TimestreamLoadResult
from timestream.record import RejectedTimestreamRecord, TimestreamRecord


class TestTimestreamLoader_UnitTests:

    @pytest.mark.unit
    def test_load(self):
        mock_session = self.get_mock_session()
        loader = TimestreamLoader('test.db', 'test.table', mock_session)
        mock_session.assert_expectations()

        result = loader.load(
            data=[],
            capture_time=datetime(2021, 1, 1, 0, 0, 0, 0, pytz.UTC)
        )
        assert type(result) == dict
        missing_keys = [
            k for k in ['error', 'invalid_items', 'accepted_records', 'rejected_records']
            if k not in result
        ]
        assert not missing_keys, f'Result {result} is missing keys {missing_keys}'
        assert type(result['invalid_items']) == list
        assert len(result['invalid_items']) == 0
        assert type(result['accepted_records']) == list
        assert len(result['accepted_records']) == 0
        assert type(result['rejected_records']) == list
        assert len(result['rejected_records']) == 0
        assert result['error'] == None
        assert loader.database_name == 'test.db'
        assert loader.table_name == 'test.table'

    @pytest.mark.unit
    def test_load_preview(self):
        mock_session = self.get_mock_session()
        loader = TimestreamLoader('test', 'test', mock_session)
        mock_session.assert_expectations()

        result = loader.load_preview(
            data=[],
            capture_time=datetime(2021, 1, 1, 0, 0, 0, 0, pytz.UTC)
        )
        assert type(result) == dict
        missing_keys = [
            k for k in ['error', 'invalid_items', 'accepted_records', 'rejected_records']
            if k not in result
        ]
        assert not missing_keys, f'Result {result} is missing keys {missing_keys}'
        assert type(result['invalid_items']) == list
        assert len(result['invalid_items']) == 0
        assert type(result['accepted_records']) == list
        assert len(result['accepted_records']) == 0
        assert type(result['rejected_records']) == list
        assert len(result['rejected_records']) == 0
        assert result['error'] == None

    @pytest.mark.unit
    def test_private_load_no_data(self):
        mock_session = self.get_mock_session()
        loader = TimestreamLoader('test', 'test', mock_session)
        mock_session.assert_expectations()

        result = loader._TimestreamLoader__load(
            data=[],
            capture_time=datetime(2021, 1, 1, 0, 0, 0, 0, pytz.UTC),
            preview=True
        )
        assert type(result) == TimestreamLoadResult
        assert len(result.invalid_items) == 0
        assert len(result.accepted_records) == 0
        assert len(result.rejected_records) == 0
        assert result.error == None

    @pytest.mark.unit
    def test_private_load_insert(self):
        # Test insert
        expected_source = 'test soure'
        expected_reason = 'test reason'
        expected_error = ClientError({
            'RejectedRecords': [{
                'RecordIndex': 1,
                'Reason': expected_reason
            }]
        }, 'test operation')
        expected_records = [{
            'MeasureName': 'valid',
            'MeasureValue': '1.0',
            'Dimensions': []
        }, {
            'MeasureName': 'invalid',
            'MeasureValue': '0.0',
            'Dimensions': []
        }, {
            'MeasureName': 'valid',
            'MeasureValue': '2.0',
            'Dimensions': [{
                'Name': 'notshared',
                'Value': 'not_shared_dimension',
                'DimensionValueType': 'VARCHAR'
            }]
        }]
        expected_common_dimensions = [{
            'Name': 'shared',
            'Value': 'shared_dimension',
            'DimensionValueType': 'VARCHAR'
        }, {
            'Name': 'source',
            'Value': 'test soure',
            'DimensionValueType': 'VARCHAR'
        }]
        expected_capture_time = '1609459200000'

        mock_client = MockApi()
        mock_client.on(
            'write_records',
            DatabaseName='test',
            TableName='test',
            CommonAttributes={
                'Dimensions': expected_common_dimensions,
                'MeasureValueType': 'DOUBLE',
                'Time': expected_capture_time,
                'Version': 1
            },
            Records=expected_records
        ).raise_exception(expected_error)
        mock_session = self.get_mock_session(mock_client)

        loader = TimestreamLoader('test', 'test', mock_session)
        result = loader._TimestreamLoader__load(
            data=[{
                'valid': 1.0,
                'invalid': 0.0,
                'shared': 'shared_dimension'
            }, {
                'valid': 2.0,
                'shared': 'shared_dimension',
                'notshared': 'not_shared_dimension'
            }],
            capture_time=datetime(2021, 1, 1, 0, 0, 0, 0, pytz.UTC),
            source=expected_source
        )
        assert type(result) == TimestreamLoadResult
        assert len(result.accepted_records) == 2
        assert result.accepted_records[0].measure_name == 'valid'
        assert len(result.rejected_records) == 1
        rejected_record = result.rejected_records[0]
        assert type(rejected_record) == RejectedTimestreamRecord
        assert str(rejected_record.error) == expected_reason
        assert rejected_record.record.measure_name == 'invalid'
        mock_session.assert_expectations()
        mock_client.assert_expectations()

    @pytest.mark.unit
    def test_private_load_upsert(self):
        # Test upsert
        expected_capture_time = '1609459200000'
        expected_records = [{
            'MeasureName': 'valid',
            'MeasureValue': '1.0',
            'Dimensions': [],
            'Time': expected_capture_time
        }]

        mock_client = MockApi()
        mock_client.on(
            'write_records',
            DatabaseName='test',
            TableName='test',
            CommonAttributes={
                'Dimensions': [],
                'MeasureValueType': 'DOUBLE',
                'Version': int(expected_capture_time)
            },
            Records=expected_records
        ).return_val(None)
        mock_session = self.get_mock_session(mock_client)

        loader = TimestreamLoader(
            'test',
            'test',
            mock_session,
            current_time_ms=int(expected_capture_time)
        )
        result = loader._TimestreamLoader__load(
            data=[{
                'valid': 1.0,
                'time': int(expected_capture_time)
            }],
            capture_time=datetime(2021, 1, 1, 0, 0, 0, 0, pytz.UTC),
            upsert=True
        )
        assert type(result) == TimestreamLoadResult
        assert len(result.accepted_records) == 1
        assert result.accepted_records[0].measure_name == 'valid'
        assert len(result.rejected_records) == 0
        mock_session.assert_expectations()
        mock_client.assert_expectations()

    @pytest.mark.unit
    def test_write_insert_records(self):
        # Set expectations
        expected_reason = 'test reason'
        expected_error = ClientError({
            'RejectedRecords': [{
                'RecordIndex': 1,
                'Reason': expected_reason
            }]
        }, 'test operation')
        expected_valid_record = TimestreamRecord('valid', 1.0)
        expected_invalid_record = TimestreamRecord('invalid', 0.0)
        expected_records = [{
            'MeasureName': 'valid',
            'MeasureValue': '1.0',
            'Dimensions': []
        }, {
            'MeasureName': 'invalid',
            'MeasureValue': '0.0',
            'Dimensions': []
        }]
        expected_capture_time = '1609459200000'

        # Set up mocks
        mock_client = MockApi()
        mock_client.on(
            'write_records',
            DatabaseName='test',
            TableName='test',
            CommonAttributes={
                'Dimensions': [],
                'MeasureValueType': 'DOUBLE',
                'Time': expected_capture_time,
                'Version': 1
            },
            Records=expected_records
        ).raise_exception(expected_error)
        mock_session = self.get_mock_session(mock_client)

        # Test ClientError occurs during Timestream write
        loader = TimestreamLoader('test', 'test', mock_session)
        accepted_records, rejected_records = loader._TimestreamLoader__write_insert_records(
            common_dimensions=TimestreamDimensions(),
            records=[expected_valid_record, expected_invalid_record],
            capture_time=datetime(2021, 1, 1, 0, 0, 0, 0, tzinfo=pytz.UTC)
        )
        assert len(accepted_records) == 1
        assert accepted_records[0] == expected_valid_record
        assert len(rejected_records) == 1

        rejected_record = rejected_records[0]
        assert type(rejected_record) == RejectedTimestreamRecord
        assert str(rejected_record.error) == expected_reason
        assert rejected_record.record == expected_invalid_record

        mock_session.assert_expectations()
        mock_client.assert_expectations()

    @pytest.mark.unit
    def test_write_upsert_records(self):
        # Set expectations
        expected_version = 1609459200000
        expected_reason = 'test reason'
        expected_error = ClientError({
            'RejectedRecords': [{
                'RecordIndex': 1,
                'Reason': expected_reason
            }]
        }, 'test operation')
        expected_valid_record = TimestreamRecord('valid', 1.0)
        expected_invalid_record = TimestreamRecord('invalid', 0.0)
        expected_records = [{
            'MeasureName': 'valid',
            'MeasureValue': '1.0',
            'Dimensions': []
        }, {
            'MeasureName': 'invalid',
            'MeasureValue': '0.0',
            'Dimensions': []
        }]

        # Set up mocks
        mock_client = MockApi()
        mock_client.on(
            'write_records',
            DatabaseName='test',
            TableName='test',
            CommonAttributes={
                'Dimensions': [],
                'MeasureValueType': 'DOUBLE',
                'Version': expected_version
            },
            Records=expected_records
        ).raise_exception(expected_error)
        mock_session = self.get_mock_session(mock_client)

        # Test ClientError occurs during Timestream write
        loader = TimestreamLoader(
            'test',
            'test',
            mock_session,
            current_time_ms=expected_version
        )
        accepted_records, rejected_records = loader._TimestreamLoader__write_upsert_records(
            common_dimensions=TimestreamDimensions(),
            records=[expected_valid_record, expected_invalid_record]
        )
        assert len(accepted_records) == 1
        assert accepted_records[0] == expected_valid_record
        assert len(rejected_records) == 1

        rejected_record = rejected_records[0]
        assert type(rejected_record) == RejectedTimestreamRecord
        assert str(rejected_record.error) == expected_reason
        assert rejected_record.record == expected_invalid_record

        mock_session.assert_expectations()
        mock_client.assert_expectations()

    @pytest.mark.unit
    def test_process_client_error(self):
        mock_session = self.get_mock_session()
        loader = TimestreamLoader('test', 'test', mock_session)
        mock_session.assert_expectations()

        expected_reason = 'test reason'
        expected_error = ClientError({
            'RejectedRecords': [{
                'RecordIndex': 1,
                'Reason': expected_reason
            }]
        }, 'test operation')
        expected_valid_record = TimestreamRecord('valid', 1.0)
        expected_invalid_record = TimestreamRecord('invalid', 0.0)

        accepted_records, rejected_records = loader._TimestreamLoader__process_client_error(
            e=expected_error,
            records=[expected_valid_record, expected_invalid_record]
        )
        assert len(accepted_records) == 1
        assert accepted_records[0] == expected_valid_record
        assert len(rejected_records) == 1

        rejected_record = rejected_records[0]
        assert type(rejected_record) == RejectedTimestreamRecord
        assert str(rejected_record.error) == expected_reason
        assert rejected_record.record == expected_invalid_record

    @pytest.mark.unit
    def test_parse_data(self):
        mock_session = self.get_mock_session()
        loader = TimestreamLoader('test', 'test', mock_session)

        # Test no data
        result = loader._TimestreamLoader__parse_data([], False, False)
        assert type(result) == TimestreamLoadResult
        assert str(result)
        assert len(result.invalid_items) == 0
        assert len(result.accepted_records) == 0
        assert len(result.rejected_records) == 0
        assert result.error == None

        # Test mix of valid and invalid items
        expected_invalid_item = {
            'unparseable': datetime(2021, 1, 1, 0, 0, 0, 0, tzinfo=pytz.UTC)
        }
        expected_error_msg = 'Some items could not be parsed into Timestream records'

        result = loader._TimestreamLoader__parse_data(
            data=[{'integer': 10}, expected_invalid_item],
            upsert=False,
            ignore_parse_error=False
        )
        assert type(result) == TimestreamLoadResult
        assert len(result.invalid_items) == 1
        assert len(result.accepted_records) == 1
        assert len(result.rejected_records) == 0
        assert str(result.error) == expected_error_msg
        assert result.invalid_items[0]['item'] == expected_invalid_item

        record = result.accepted_records[0]
        assert record and type(record) == TimestreamRecord
        assert record.measure_name == 'integer'
        assert record.measure_value == 10.0
        assert record.mstime == None
        assert record.time == None

    @pytest.mark.unit
    def test_parse_item(self):
        mock_session = self.get_mock_session()
        loader = TimestreamLoader('test', 'test', mock_session)
        mock_session.assert_expectations()

        # Test no item
        assert loader._TimestreamLoader__parse_item({}, False, False) == []

        # Test parse insert item
        expected_dimensions = TimestreamDimensions({
            'string': 'dimension',
            'remove this': 'dimension'
        })
        expected_dimensions.remove_dimension('remove this')
        records = loader._TimestreamLoader__parse_item({
            'integer': 10,
            'float': 20.0,
            'list': [10, 20.0],
            'string': 'dimension',
            'none': None
        }, False, False)
        assert len(records) == 3
        assert str(records[0])

        record = records[0]
        assert record and type(record) == TimestreamRecord
        assert record.measure_name == 'integer'
        assert record.measure_value == 10.0
        assert record.mstime == None
        assert record.time == None
        assert record.dimensions == expected_dimensions
        assert record.dimensions.get_dimension('string') == 'dimension'
        assert str(record.dimensions)

        record = records[1]
        assert record and type(record) == TimestreamRecord
        assert record.measure_name == 'float'
        assert record.measure_value == 20.0
        assert record.mstime == None
        assert record.time == None
        assert record.dimensions == expected_dimensions

        record = records[2]
        assert record and type(record) == TimestreamRecord
        assert record.measure_name == 'list'
        assert record.measure_value == 2.0
        assert record.mstime == None
        assert record.time == None
        assert record.dimensions == expected_dimensions

        # Test parse upsert item
        records = loader._TimestreamLoader__parse_item({
            'Time': 1609459200000,
            'integer': 10
        }, True, False)
        assert len(records) == 1

        record = records[0]
        assert record and type(record) == TimestreamRecord
        assert record.measure_name == 'integer'
        assert record.measure_value == 10.0
        assert record.mstime == 1609459200000
        assert record.time == datetime.combine(
            date(2021, 1, 1),
            datetime.min.time()
        ).replace(tzinfo=pytz.UTC)

        # Test parse error
        with pytest.raises(AssertionError) as e_info:
            loader._TimestreamLoader__parse_item({
                'unparseable': {'key': 'value'}
            }, False, False)
        assert str(e_info.value) == 'Could not create a record or dimension from type "<class \'dict\'>" at key unparseable'

        # Test ignore parse error
        records = loader._TimestreamLoader__parse_item({
            'unparseable': {'key': 'value'}
        }, False, True)
        assert len(records) == 0

    @pytest.mark.unit
    def test_does_database_exist(self):
        # Set expectations
        mock_response = {'Database': {'Arn': 'arn:aws:timestream:us-east-1:000000000000:database/test', 'DatabaseName': 'test', 'TableCount': 1, 'KmsKeyId': 'arn:aws:kms:us-east-1:000000000000:key/12345678-1234-1234-1234-1234567890ab', 'CreationTime': datetime(2021, 7, 19, 16, 25, 26, 419000), 'LastUpdatedTime': datetime(2021, 10, 8, 16, 25, 37, 893000)}, 'ResponseMetadata': {'RequestId': '1234567890abcdefghijklmnop', 'HTTPStatusCode': 200, 'HTTPHeaders': {'x-amzn-requestid': '1234567890abcdefghijklmnop', 'content-type': 'application/x-amz-json-1.0', 'content-length': '289', 'date': 'Wed, 20 Oct 2021 16:54:48 GMT'}, 'RetryAttempts': 0}}

        # Set up mocks
        mock_client = MockApi()
        mock_client.on(
            'describe_database',
            DatabaseName='test',
        ).return_val(mock_response)
        mock_session = self.get_mock_session(mock_client)

        # Test True when using TimestreamLoader's database attribute
        loader = TimestreamLoader('test', 'test', mock_session)
        result = loader.does_database_exist()
        assert result
        mock_session.assert_expectations()
        mock_client.assert_expectations()

    @pytest.mark.unit
    def test_does_table_exist(self):
        # Set expectations
        mock_response = {'Table': {'Arn': 'arn:aws:timestream:us-east-1:000000000000:database/test/table/test', 'TableName': 'test', 'DatabaseName': 'test', 'TableStatus': 'ACTIVE', 'RetentionProperties': {'MemoryStoreRetentionPeriodInHours': 720, 'MagneticStoreRetentionPeriodInDays': 1}, 'CreationTime': datetime(2021, 8, 19, 18, 21, 1, 519000), 'LastUpdatedTime': datetime(2021, 8, 31, 12, 36, 4, 503000)}, 'ResponseMetadata': {'RequestId': 'SDKOJJS7STGCQXOFVLXPBYDRKI', 'HTTPStatusCode': 200, 'HTTPHeaders': {'x-amzn-requestid': 'SDKOJJS7STGCQXOFVLXPBYDRKI', 'content-type': 'application/x-amz-json-1.0', 'content-length': '366', 'date': 'Wed, 20 Oct 2021 17:41:09 GMT'}, 'RetryAttempts': 1}}

        # Set up mocks
        mock_client = MockApi()
        mock_client.on(
            'describe_table',
            DatabaseName='test',
            TableName='test'
        ).return_val(mock_response)
        mock_session = self.get_mock_session(mock_client)

        # Test True when using TimestreamLoader's database and table attribute
        loader = TimestreamLoader('test', 'test', mock_session)
        result = loader.does_table_exist()
        assert result
        mock_session.assert_expectations()
        mock_client.assert_expectations()

    def get_mock_session(self, write_client=None) -> MockApi:
        mock_session = MockApi()
        mock_session.on(
            'client',
            'timestream-write',
            config=DEFAULT_WRITE_CLIENT_CONFIG
        ).return_val(write_client)
        mock_session.on(
            'client',
            'timestream-query'
        ).return_val(MockApi())
        return mock_session
